#!/bin/bash
########################################################################
# build_hepmc3.sh
########################################################################

####################################
# Useful variables.
prog=${0##*/}
thisdir=$(cd "$(dirname "${0}")" && pwd -P)
trace=maybe_xtrace # Could set to xtrace instead for unconditional.

####################################
# Useful functions.

# Report usage.
usage() {
  cat <<EOF
USAGE: ${prog} ${ssi_msg_config_usage:-<ssi-msg-config-options>} <product-dir> (e19|e20|c7):p382 debug|prof [tar]
       ${prog} -h|--help

  Build ${package} and package for UPS.

EOF
  (( ${1:-0} < 2 )) && return;
  cat <<EOF
OPTIONS

  -h|--help

     This help.

${ssi_msg_config_options:-"  <ssi-msg-config-options>

    [Supply <product-dir> to enable more details]

"}\
ARGUMENTS

  <product-dir>
    The top-level UPS directory in which the source code distribution
    should be constructed.

  (e19|e20|c7):p382
    Qualifier string.

  debug|prof
    Build type.

  [tar]
    Generate a binary installation archive.

${ssi_msg_config_environment}
ENVIRONMENT

  DEBUG_BUILD
    Full command tracing for the build (VERBOSE=1)

  DEBUG_PATCHES
  VERBOSE_PATCHES
    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.

${ssi_msg_config_environment}
EOF
}

# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}

# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    # We\'ve been asked for help: give as much as we can.
    usage 2; exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> ${product_dir}\n$(usage)"
    exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: ${product_dir}"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}"
  # Absolute is safer.
  product_dir="${abs_product_dir}"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
    source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "${product_dir}" && \
    setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
    ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
}

# Handle qualifiers.
set_qualifiers() {
  local qlist=(`echo ${basequal} | sed -e 's/:/ /g' | tr " " "\n"| sort |tr "\n" " "`)
  local qual
  for qual in "${qlist[@]}"; do
    case $qual in
      e19|e20|c7) cqual=$qual;;
      p382) pqual=$qual;;
      *) ssi_die "Unrecognized qualifier $qual"
    esac
  done
  [[ -n "$cqual" ]] || ssi_die "Base qualifier is mandatory"
  [[ -n "$pqual" ]] || ssi_die "A python qualifier is mandatory (e.g. p382)"
  basequal=${cqual}:${pqual}
}

# Deal with compiler-specific options.
add_compiler_options() {
  local cc cxx cxxstd
  case ${cqual} in
    e19|e20)
      cc=gcc
      cxx=g++
      cxxstd=17
      ;;
    c7)
      cc=clang
      cxx=clang++
      cxxstd=17
      ;;
    *)
      ssi_die "unrecognized compiler qualifier ${cqual}\n$(usage)"
  esac
  cmake_options+=( \
    -DCMAKE_C_COMPILER=${cc} \
    -DCMAKE_C_FLAGS_RELEASE="-O3 -DNDEBUG" \
    -DCMAKE_C_FLAGS_RELWITHDEBINFO="-g -O3 -DNDEBUG -fno-omit-frame-pointer" \
    -DCMAKE_C_FLAGS_DEBUG="-g -O0" \
    -DCMAKE_CXX_COMPILER=${cxx} \
    -DCMAKE_CXX_FLAGS_RELEASE="-O3 -DNDEBUG" \
    -DCMAKE_CXX_FLAGS_RELWITHDEBINFO="-g -O3 -DNDEBUG -fno-omit-frame-pointer -Wall -Wextra -pedantic" \
    -DCMAKE_CXX_FLAGS_DEBUG="-g -O0" \
    -DCMAKE_CXX_STANDARD=${cxxstd} \
    -DCMAKE_CXX_EXTENSIONS=OFF \
    )
}

# Ascertain whether we built and installed correctly.
is_built() {
  verify_in_filesystem -d "${@}" "${pkgdir}/lib"{64,} 
}

# If a distribution tarball is wanted and the installation is
# successful, then make it and exit.
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
    is_built "$@" && make_basic_distribution && exit
  fi
}

# Determine the CMake build type from the UPS build type.
cmake_build_type() {
  local ups_build_type="${1}"
  local cmake_build_type
  case ${ups_build_type} in
    debug) cmake_build_type=Debug;;
    prof) cmake_build_type=RelWithDebInfo;;
    opt) cmake_build_type=Release;;
    *) ssi_die -I "please specify debug, opt, or prof\n$(usage)"
  esac
  echo "${cmake_build_type}"
}

########################################################################
# Main.
########################################################################

####################################
# Package name and version information.
package=hepmc3
origpkgver=v3_2_1
pkgver=${origpkgver}
ssibuildshims_version=v2_00_00
pkgdotver=${origpkgver//_/.}; pkgdotver=${pkgdotver#v}
pkgtardir=HepMC3-${pkgdotver}
pkgtarfile=${pkgtardir}.tar.gz

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

# Initialization and defaults.
(( ssi_trace = 1 ))

# Parse options.
while true; do
  case $1 in
    -h|--help) usage 2; exit 1;;
    --) shift; break;;
    -*) ssi_handle_config_options "${1}" || \
      ssi_die 2 "unrecognized option $1\n$(usage)";;
    *) shift; break # Swallow <product-dir>: dealt with.
  esac
  shift
done

# Environment overrides.
ssi_honor_environment_config

##################
# Remaining arguments.
basequal=${1}; shift
extraqual=${1}; shift
maketar=${1}; shift

# Deal with qualifiers.
set_qualifiers

# Define globals.
define_basics || exit

# Possible short-circuit.
maybe_make_distribution -q

ssi_info "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${pkgflvr})"

####################################
# Ensure directories.
mkdir -p "${pkgdir}" || \
  ssi_die "failed to create installation directory ${pkgdir}"
# Remove any existing sourcedir to avoid patch conflicts.
rm -rf "${sourcedir}" && mkdir -p "${sourcedir}" || \
  ssi_die "failed to create source directory ${sourcedir}"
# Clean build.
rm -rf "${blddir}" && mkdir -p "${blddir}" || \
  ssi_die "failed to create build directory ${blddir}"
####################################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies.
${trace} fake_declare_product "${ssi_args[@]}" -x \
  "${product_dir}" ${package} ${pkgver} ${fullqual} || exit
####################################

####################################
# Build-only dependencies.

# Require CMake.
${trace} setup_cmake --latest  || exit

# We can use GNU Make, but we'll take Ninja if it's available
${trace} setup_ninja --optional && (( want_ninja = 1 ))
####################################

# Print active products.
echo
ups active
echo

cd "${sourcedir}" || ssi_die "unable to cd to source directory ${sourcedir}"

# Expand and patch the source.
${trace} prepare_source "${tardir}/${pkgtarfile}" \
  fix-HepMC3-config.patch || exit

[ -n "${DEBUG_PATCHES}" ] && exit 1

####################################
# Set all CMake options.

# Start with generator, if appropriate.
declare -a cmake_options
(( want_ninja )) && cmake_options+=(-GNinja)

if [[ "${PYTHON_VERSION}" =~ ^v([0-9]+)_([0-9]+) ]]; then
  pythonMm="${BASH_REMATCH[1]}${BASH_REMATCH[2]}"
  pythonMdotm="${BASH_REMATCH[1]}.${BASH_REMATCH[2]}"
else
  ssi_die "unable to extract major and minor versions from Python version ${PYTHON_VERSION}"
fi

# Note we don't know exactly where HEPMC3_LIB is going to be (thanks to
# GNUInstallDirs.cmake and Debian, etc.) so we'll find out later and relocate.
tmp_python_installdir="${pkgdir}/${PYTHON_LIBDIR}/site-packages"

# Set main CMake options.
cmake_options=( \
  -DCMAKE_INSTALL_PREFIX=${pkgdir} \
  -DCMAKE_BUILD_TYPE=$(cmake_build_type "${extraqual}") \
  -DHEPMC3_ENABLE_SEARCH:BOOL=OFF \
  -DHEPMC3_ENABLE_ROOTIO:BOOL=OFF \
  -DHEPMC3_ENABLE_PYTHON:BOOL=ON \
  -DHEPMC3_PYTHON_VERSIONS="${pythonMdotm}" \
  -DHEPMC3_Python_SITEARCH${pythonMm}="${tmp_python_installdir}" \
  -DHEPMC3_BUILD_DOCS:BOOL=OFF \
  -DHEPMC3_BUILD_EXAMPLES:BOOL=OFF \
  -DHEPMC3_INSTALL_INTERFACES:BOOL=ON \
  -DHEPMC3_ENABLE_TEST:BOOL=ON \
  )

# Set compiler-specific options.
add_compiler_options
####################################

####################################
# Ascertain and configure parallelism.
ncore=$(ncores)

# Default parallelism if not already set. 
[[ -n "${CMAKE_BUILD_PARALLEL_LEVEL}" ]] || \
  export CMAKE_BUILD_PARALLEL_LEVEL=${ncore}

[[ -n "${CTEST_PARALLEL_LEVEL}" ]] || \
  export CTEST_PARALLEL_LEVEL=${ncore}
####################################

cd "${blddir}" || ssi_die "unable to cd to build directory ${blddir}"

####################################
# Configure, build, and install.

# Configure.
${trace} cmake "${cmake_options[@]}" "${sourcedir}/${pkgtardir}" || \
  ssi_die "configure failed"

# Build.
${trace} cmake --build "${blddir}" ${DEBUG_BUILD:+--verbose} || \
  ssi_die "build failed"

# Test.
${trace} ctest || ssi_die "tests failed"

# Install.
${trace} cmake --install "${blddir}" || ssi_die "install failed"

####################################
# Installation verification and post-processing.

# Verify installation.

# This check is not part of is_built because the manifest is a
# non-installed build artifact required only during installation
# post-processing.
manifest="${blddir}/install_manifest.txt"
verify_in_filesystem -s "${manifest}" || \
  ssi_build_failed ${package} "required build artifact ${manifest} missing"

is_built -q || ssi_build_failed ${package}

# Ascertain libdir from the manifest.
libdir=($(cat "${manifest}" | \
  while read file; do basename "$(dirname "${file}")"; done | \
  grep -e '^lib' | head -1))
(( ${#libdir[@]} == 1 )) || \
  ssi_die "unable to ascertain libdir from CMake installation manifest (${libdir[@]})"
verify_in_filesystem -d "${pkgdir}/${libdir}" || \
  ssi_build_failed ${package} \
  "$libdir is mentioned in $manifest but missing in filesystem"

# Relocate the python directory to its final resting place.
if [[ -d "${tmp_python_installdir}" ]]; then
  mv -v "${tmp_python_installdir%/*}" "${pkgdir}/${libdir}/" || \
    ssi_die "unable to move Python directory into ${pkgdir}/${libdir}"
fi

# Fill in libdir in arch-specific UPS table file.
mkdir -p "${pkgdir}/ups" || ssi_die "unable to create arch-specific UPS table directory"
perl -wpe 's&\@LIBDIR\@&'"$libdir"'&g' "${pkgups}/${package}.table" > \
  "${pkgdir}/ups/${package}.table" && \
  [[ -s "${pkgdir}/ups/${package}.table" ]] || \
  ssi_die "unable to create arch-specific UPS table file"
####################################

ssi_info "finished building ${package} ${pkgver}"

# Declare and set up the installed package.
unsetup ${package} >/dev/null 2>&1
${trace} declare_product -x -M "$(basename "${pkgdir}")"/ups "${ssi_args[@]}" \
  "${product_dir}" ${package} ${pkgver} ${fullqual} || exit

ssi_info "${package} is installed at ${pkgdir}"

# This must be last.
maybe_make_distribution
